// The MIT License (MIT)
//
// Copyright (c) 2020 The Prometheus Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
import { snippet } from '@nexucis/codemirror-next-autocomplete';
import { AggregateOp, BinaryExpr, FunctionIdentifier, GroupingLabel, GroupingLabels, Identifier, LabelMatcher, LabelMatchers, LabelName, MetricIdentifier, StringLiteral, VectorSelector, MatchOp, } from 'lezer-promql';
import { walkBackward, walkThrough } from '../parser/path-finder';
import { aggregateOpModifierTerms, aggregateOpTerms, binOpModifierTerms, binOpTerms, functionIdentifierTerms, matchOpTerms } from './promql.terms';
const autocompleteNode = {
    matchOp: { labels: matchOpTerms, type: '' },
    binOp: { labels: binOpTerms, type: '' },
    binOpModifier: { labels: binOpModifierTerms, type: 'keyword' },
    functionIdentifier: { labels: functionIdentifierTerms, type: 'function' },
    aggregateOp: { labels: aggregateOpTerms, type: 'keyword' },
    aggregateOpModifier: { labels: aggregateOpModifierTerms, type: 'keyword' },
};
const snippets = [
    {
        keyword: 'sum(rate(__input_vector__[5m]))',
        snippet: 'sum(rate(${__input_vector__}[5m]))',
    },
    {
        keyword: 'histogram_quantile(__quantile__, sum by(le) (rate(__histogram_metric__[5m])))',
        snippet: 'histogram_quantile(${__quantile__}, sum by(le) (rate(${__histogram_metric__}[5m])))',
    },
    {
        keyword: 'label_replace(__input_vector__, "__dst__", "__replacement__", "__src__", "__regex__")',
        snippet: 'label_replace(${__input_vector__}, "${__dst__}", "${__replacement__}", "${__src__}", "${__regex__}")',
    },
];
// parsedSnippets shouldn't be modified. It's only there to not parse everytime the above list of snippet
const parsedSnippets = snippets.map((s) => ({
    label: s.name || s.keyword,
    original: s.name || s.keyword,
    apply: snippet(s.snippet),
    score: 0,
}));
// HybridComplete provides a full completion result with or without a remote prometheus.
export class HybridComplete {
    constructor(prometheusClient) {
        this.prometheusClient = prometheusClient;
    }
    promQL(context) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s;
        const { state, pos } = context;
        const tree = state.tree.resolve(pos, -1);
        if (((_a = tree.parent) === null || _a === void 0 ? void 0 : _a.type.id) === MetricIdentifier && tree.type.id === Identifier) {
            let nonMetricCompletions = [autocompleteNode.functionIdentifier, autocompleteNode.aggregateOp, autocompleteNode.binOp];
            if (((_e = (_d = (_c = (_b = tree.parent) === null || _b === void 0 ? void 0 : _b.parent) === null || _c === void 0 ? void 0 : _c.parent) === null || _d === void 0 ? void 0 : _d.parent) === null || _e === void 0 ? void 0 : _e.type.id) === BinaryExpr) {
                // This is for autocompleting binary operator modifiers (on / ignoring / group_x). When we get here, we have something like:
                //       metric_name / ignor
                // And the tree components above the half-finished set operator will look like:
                //
                // Identifier -> MetricIdentifier -> VectorSelector -> Expr -> BinaryExpr.
                nonMetricCompletions = nonMetricCompletions.concat(autocompleteNode.binOpModifier);
            }
            // Here we cannot know if we have to autocomplete the metric_name, or the function or the aggregation.
            // So we will just autocomplete everything
            if (this.prometheusClient) {
                return this.prometheusClient.labelValues('__name__').then((metricNames) => {
                    const result = [{ labels: metricNames, type: 'constant' }];
                    return this.arrayToCompletionResult(result.concat(nonMetricCompletions), tree.start, pos, context, state, true);
                });
            }
            return this.arrayToCompletionResult(nonMetricCompletions, tree.start, pos, context, state, true);
        }
        if (tree.type.id === GroupingLabels || (((_f = tree.parent) === null || _f === void 0 ? void 0 : _f.type.id) === GroupingLabel && tree.type.id === LabelName)) {
            // In this case we are in the given situation:
            //      sum by ()
            // So we have to autocomplete any labelName
            return this.labelNames(tree, pos, context, state);
        }
        if (tree.type.id === LabelMatchers || (((_g = tree.parent) === null || _g === void 0 ? void 0 : _g.type.id) === LabelMatcher && tree.type.id === LabelName)) {
            // In that case we are in the given situation:
            //       metric_name{} or {}
            return this.autocompleteLabelNamesByMetric(tree, pos, context, state);
        }
        if (((_h = tree.parent) === null || _h === void 0 ? void 0 : _h.type.id) === LabelMatcher && tree.type.id === StringLiteral) {
            // In this case we are in the given situation:
            //      metric_name{labelName=""}
            // So we can autocomplete the labelValue
            return this.autocompleteLabelValue(tree.parent, tree, pos, context, state);
        }
        if (tree.type.id === LabelMatcher &&
            ((_j = tree.firstChild) === null || _j === void 0 ? void 0 : _j.type.id) === LabelName &&
            ((_k = tree.lastChild) === null || _k === void 0 ? void 0 : _k.type.id) === 0 &&
            ((_l = tree.lastChild) === null || _l === void 0 ? void 0 : _l.firstChild) === null // Discontinues completion in invalid cases like `foo{bar==<cursor>}`
        ) {
            // In this case the current token is not itself a valid match op yet:
            //      metric_name{labelName!}
            return this.arrayToCompletionResult([autocompleteNode.matchOp], tree.lastChild.start, pos, context, state);
        }
        if (tree.type.id === MatchOp || ((_m = tree.parent) === null || _m === void 0 ? void 0 : _m.type.id) === MatchOp) {
            // In this case the current token is already a valid match op, but could be extended, e.g. "=" to "=~".
            return this.arrayToCompletionResult([autocompleteNode.matchOp], tree.start, pos, context, state);
        }
        if (((_o = tree.parent) === null || _o === void 0 ? void 0 : _o.type.id) === BinaryExpr) {
            return this.arrayToCompletionResult([autocompleteNode.binOp], tree.start, pos, context, state);
        }
        if (((_p = tree.parent) === null || _p === void 0 ? void 0 : _p.type.id) === FunctionIdentifier) {
            return this.arrayToCompletionResult([autocompleteNode.functionIdentifier], tree.start, pos, context, state);
        }
        if (((_q = tree.parent) === null || _q === void 0 ? void 0 : _q.type.id) === AggregateOp) {
            return this.arrayToCompletionResult([autocompleteNode.aggregateOp], tree.start, pos, context, state);
        }
        if ((tree.type.id === Identifier && ((_r = tree.parent) === null || _r === void 0 ? void 0 : _r.type.id) === 0) || (tree.type.id === 0 && ((_s = tree.parent) === null || _s === void 0 ? void 0 : _s.type.id) !== LabelMatchers)) {
            // This matches identifier-ish keywords in certain places where a normal identifier would be invalid, like completing "b" into "by":
            //        sum b
            // ...or:
            //        sum(metric_name) b
            // ...or completing "unle" into "unless":
            //        metric_name / unle
            // TODO: This is imprecise and autocompletes in too many situations. Make this better.
            return this.arrayToCompletionResult([autocompleteNode.aggregateOpModifier].concat(autocompleteNode.binOp), tree.start, pos, context, state);
        }
        return null;
    }
    autocompleteLabelValue(parent, current, pos, context, state) {
        if (!this.prometheusClient) {
            return null;
        }
        // First get the labelName.
        // By definition it's the firstChild: https://github.com/promlabs/lezer-promql/blob/0ef65e196a8db6a989ff3877d57fd0447d70e971/src/promql.grammar#L250
        let labelName = '';
        if (this.prometheusClient && parent.firstChild && parent.firstChild.type.id === LabelName) {
            labelName = state.sliceDoc(parent.firstChild.start, parent.firstChild.end);
        }
        // then find the metricName if it exists
        const metricName = this.getMetricNameInVectorSelector(current, state);
        return this.prometheusClient.labelValues(labelName, metricName).then((labelValues) => {
            // +1 to avoid to remove the first quote.
            return this.arrayToCompletionResult([
                {
                    labels: labelValues,
                    type: 'text',
                },
            ], current.start + 1, pos, context, state);
        });
    }
    autocompleteLabelNamesByMetric(tree, pos, context, state) {
        return this.labelNames(tree, pos, context, state, this.getMetricNameInVectorSelector(tree, state));
    }
    getMetricNameInVectorSelector(tree, state) {
        // Find if there is a defined metric name. Should be used to autocomplete a labelValue or a labelName
        // First find the parent "VectorSelector" to be able to find then the subChild "MetricIdentifier" if it exists.
        let currentNode = walkBackward(tree, VectorSelector);
        if (!currentNode) {
            // Weird case that shouldn't happen, because "VectorSelector" is by definition the parent of the LabelMatchers.
            return '';
        }
        currentNode = walkThrough(currentNode, MetricIdentifier, Identifier);
        if (!currentNode) {
            return '';
        }
        return state.sliceDoc(currentNode.start, currentNode.end);
    }
    labelNames(tree, pos, context, state, metricName) {
        return !this.prometheusClient
            ? null
            : this.prometheusClient.labelNames(metricName).then((labelNames) => {
                // this case can happen when you are in empty bracket. Then you don't want to remove the first bracket
                return this.arrayToCompletionResult([
                    {
                        labels: labelNames,
                        type: 'constant',
                    },
                ], tree.type.id === GroupingLabels || tree.type.id === LabelMatchers ? tree.start + 1 : tree.start, pos, context, state);
            });
    }
    arrayToCompletionResult(data, from, to, context, state, includeSnippet = false) {
        const text = state.sliceDoc(from, to);
        const options = [];
        for (const completionList of data) {
            for (const label of completionList.labels) {
                const completionResult = context.filter({
                    label: label,
                    original: label,
                    apply: label,
                    type: completionList.type,
                    score: 0,
                }, text);
                if (completionResult !== null) {
                    options.push(completionResult);
                }
            }
        }
        if (includeSnippet) {
            for (const s of parsedSnippets) {
                const completionResult = context.filter(s, text);
                if (completionResult !== null) {
                    options.push(completionResult);
                }
            }
        }
        return {
            from: from,
            to: to,
            options: options,
            filterDownOn: /^[a-zA-Z0-9_:]+$/,
        };
    }
}
//# sourceMappingURL=hybrid.js.map